@@ -7,13 +7,12 @@ if 'SUMO_HOME' in os.environ:
     sys.path.append(tools)
 else:
     sys.exit("Please declare the environment variable 'SUMO_HOME'")
-import gym
+import gymnasium as gym
 import numpy as np
 import pandas as pd
 import sumolib
 import traci
-from gym.envs.registration import EnvSpec
-from gym.utils import EzPickle, seeding
+from gymnasium.utils import EzPickle, seeding
 from pettingzoo import AECEnv
 from pettingzoo.utils import agent_selector, wrappers
 from pettingzoo.utils.agent_selector import agent_selector
@@ -50,14 +49,20 @@ class SumoEnvironment(gym.Env):
     :param min_green: (int) Minimum green time in a phase
     :param max_green: (int) Max green time in a phase
     :single_agent: (bool) If true, it behaves like a regular gym.Env. Else, it behaves like a MultiagentEnv (https://github.com/ray-project/ray/blob/master/python/ray/rllib/env/multi_agent_env.py)
-    :reward_fn: (str/function) String with the name of the reward function used by the agents, or a reward function.
+    :reward_fn: (str/function/dict) String with the name of the reward function used by the agents, a reward function, or dictionary with reward functions assigned to individual traffic lights by their keys
+    :observation_fn: (str/function) String with the name of the observation function or a callable observation function itself
     :add_system_info: (bool) If true, it computes system metrics (total queue, total waiting time, average speed) in the info dictionary
     :add_per_agent_info: (bool) If true, it computes per-agent (per-traffic signal) metrics (average accumulated waiting time, average queue) in the info dictionary
     :sumo_seed: (int/string) Random seed for sumo. If 'random' it uses a randomly chosen seed.
     :fixed_ts: (bool) If true, it will follow the phase configuration in the route_file and ignore the actions.
     :sumo_warnings: (bool) If False, remove SUMO warnings in the terminal
     :additional_sumo_cmd: (list) Additional command line arguments for SUMO
+    :render_mode: (str) Mode of rendering. Can be 'human' or 'rgb_array'
     """
+    metadata = {
+        'render_modes': ['human', 'rgb_array'],
+    }
+
     CONNECTION_LABEL = 0  # For traci multi-client support
 
     def __init__(
@@ -66,7 +71,7 @@ class SumoEnvironment(gym.Env):
         route_file: str, 
         out_csv_name: Optional[str] = None,
         use_gui: bool = False, 
-        virtual_display: Optional[Tuple[int,int]] = None,
+        virtual_display: Tuple[int,int] = (3200, 1800),
         begin_time: int = 0, 
         num_seconds: int = 20000, 
         max_depart_delay: int = 100000,
@@ -77,24 +82,29 @@ class SumoEnvironment(gym.Env):
         min_green: int = 5, 
         max_green: int = 50, 
         single_agent: bool = False,
-        reward_fn: Union[str,Callable] = 'diff-waiting-time',
+        reward_fn: Union[str,Callable,dict] = 'diff-waiting-time',
+        observation_fn: Union[str,Callable] = 'default',
         add_system_info: bool = True,
         add_per_agent_info: bool = True,
         sumo_seed: Union[str,int] = 'random', 
         fixed_ts: bool = False,
         sumo_warnings: bool = True,
         additional_sumo_cmd: Optional[str] = None,
-    ):
+        render_mode: Optional[str] = None,
+    ) -> None:
+        assert render_mode is None or render_mode in self.metadata["render_modes"], "Invalid render mode."
+        self.render_mode = render_mode  
+        self.virtual_display = virtual_display
+        self.disp = None
+
         self._net = net_file
         self._route = route_file
         self.use_gui = use_gui
-        if self.use_gui:
+        if self.use_gui or self.render_mode is not None:
             self._sumo_binary = sumolib.checkBinary('sumo-gui')
         else:
             self._sumo_binary = sumolib.checkBinary('sumo')
 
-        self.virtual_display = virtual_display
-
         assert delta_time > yellow_time, "Time between actions must be at least greater than yellow time."
 
         self.begin_time = begin_time
@@ -125,21 +135,34 @@ class SumoEnvironment(gym.Env):
             traci.start([sumolib.checkBinary('sumo'), '-n', self._net], label='init_connection'+self.label)
             conn = traci.getConnection('init_connection'+self.label)
         self.ts_ids = list(conn.trafficlight.getIDList())
-        self.traffic_signals = {ts: TrafficSignal(self, 
-                                                  ts, 
-                                                  self.delta_time, 
-                                                  self.yellow_time, 
-                                                  self.min_green, 
-                                                  self.max_green, 
-                                                  self.begin_time,
-                                                  self.reward_fn,
-                                                  conn) for ts in self.ts_ids}
+        self.observation_fn = observation_fn
+
+        if isinstance(self.reward_fn, dict):
+            self.traffic_signals = {ts: TrafficSignal(self,
+                                                      ts,
+                                                      self.delta_time,
+                                                      self.yellow_time,
+                                                      self.min_green,
+                                                      self.max_green,
+                                                      self.begin_time,
+                                                      self.reward_fn[ts],
+                                                      conn) for ts in self.reward_fn.keys()}
+        else:
+            self.traffic_signals = {ts: TrafficSignal(self,
+                                                      ts,
+                                                      self.delta_time,
+                                                      self.yellow_time,
+                                                      self.min_green,
+                                                      self.max_green,
+                                                      self.begin_time,
+                                                      self.reward_fn,
+                                                      conn) for ts in self.ts_ids}
+
         conn.close()
 
         self.vehicles = dict()
         self.reward_range = (-float('inf'), float('inf'))
         self.metadata = {}
-        self.spec = EnvSpec('SUMORL-v0')
         self.run = 0
         self.metrics = []
         self.out_csv_name = out_csv_name
@@ -163,9 +186,9 @@ class SumoEnvironment(gym.Env):
             sumo_cmd.append('--no-warnings')
         if self.additional_sumo_cmd is not None:
             sumo_cmd.extend(self.additional_sumo_cmd.split())
-        if self.use_gui:
+        if self.use_gui or self.render_mode is not None:
             sumo_cmd.extend(['--start', '--quit-on-end'])
-            if self.virtual_display is not None:
+            if self.render_mode == 'rgb_array':
                 sumo_cmd.extend(['--window-size', f'{self.virtual_display[0]},{self.virtual_display[1]}'])
                 from pyvirtualdisplay.smartdisplay import SmartDisplay
                 print("Creating a virtual display.")
@@ -180,11 +203,11 @@ class SumoEnvironment(gym.Env):
             traci.start(sumo_cmd, label=self.label)
             self.sumo = traci.getConnection(self.label)
         
-        if self.use_gui:
+        if self.use_gui or self.render_mode is not None:
             self.sumo.gui.setSchema(traci.gui.DEFAULT_VIEW, "real world")                
 
-    def reset(self, seed: Optional[int] = None, return_info=False, **kwargs):
-        super().reset(seed=seed, return_info=return_info, **kwargs)
+    def reset(self, seed: Optional[int] = None, **kwargs):
+        super().reset(seed=seed, **kwargs)
         
         if self.run != 0:
             self.close()
@@ -196,22 +219,31 @@ class SumoEnvironment(gym.Env):
             self.sumo_seed = seed
         self._start_simulation()
 
-        self.traffic_signals = {ts: TrafficSignal(self, 
-                                                  ts, 
-                                                  self.delta_time, 
-                                                  self.yellow_time, 
-                                                  self.min_green, 
-                                                  self.max_green, 
-                                                  self.begin_time,
-                                                  self.reward_fn,
-                                                  self.sumo) for ts in self.ts_ids}
+        if isinstance(self.reward_fn, dict):
+            self.traffic_signals = {ts: TrafficSignal(self,
+                                                      ts,
+                                                      self.delta_time,
+                                                      self.yellow_time,
+                                                      self.min_green,
+                                                      self.max_green,
+                                                      self.begin_time,
+                                                      self.reward_fn[ts],
+                                                      self.sumo) for ts in self.reward_fn.keys()}
+        else:
+            self.traffic_signals = {ts: TrafficSignal(self,
+                                                      ts,
+                                                      self.delta_time,
+                                                      self.yellow_time,
+                                                      self.min_green,
+                                                      self.max_green,
+                                                      self.begin_time,
+                                                      self.reward_fn,
+                                                      self.sumo) for ts in self.ts_ids}
+
         self.vehicles = dict()
 
         if self.single_agent:
-            if return_info:
-                return self._compute_observations()[self.ts_ids[0]], self._compute_info()
-            else:
-                return self._compute_observations()[self.ts_ids[0]]
+            return self._compute_observations()[self.ts_ids[0]], self._compute_info()
         else:
             return self._compute_observations()
 
@@ -234,10 +266,12 @@ class SumoEnvironment(gym.Env):
         observations = self._compute_observations()
         rewards = self._compute_rewards()
         dones = self._compute_dones()
+        terminated = False  # there are no 'terminal' states in this environment
+        truncated = dones['__all__']  # episode ends when sim_step >= max_steps
         info = self._compute_info()
 
         if self.single_agent:
-            return observations[self.ts_ids[0]], rewards[self.ts_ids[0]], dones['__all__'], info
+            return observations[self.ts_ids[0]], rewards[self.ts_ids[0]], terminated, truncated, info
         else:
             return observations, rewards, dones, info
 
@@ -334,25 +368,24 @@ class SumoEnvironment(gym.Env):
         if not LIBSUMO:
             traci.switch(self.label)
         traci.close()
-        try:
+        if self.disp is not None:
             self.disp.stop()
-        except AttributeError:
-            pass
+            self.disp = None
         self.sumo = None
     
     def __del__(self):
-        self.close()
-    
-    def render(self, mode='human'):
-        if self.virtual_display:
+        self.close()        
+
+    def render(self):
+        if self.render_mode == 'human':
+            return # sumo-gui will already be rendering the frame
+        elif self.render_mode == 'rgb_array':
             #img = self.sumo.gui.screenshot(traci.gui.DEFAULT_VIEW,
             #                          f"temp/img{self.sim_step}.jpg", 
             #                          width=self.virtual_display[0],
             #                          height=self.virtual_display[1])
             img = self.disp.grab()
-            if mode == 'rgb_array':
-                return np.array(img)
-            return img         
+            return np.array(img)                
     
     def save_csv(self, out_csv_name, run):
         if out_csv_name is not None:
@@ -393,19 +426,21 @@ class SumoEnvironmentPZ(AECEnv, EzPickle):
 
         # dicts
         self.rewards = {a: 0 for a in self.agents}
-        self.dones = {a: False for a in self.agents}
+        self.terminations = {a: False for a in self.agents}
+        self.truncations = {a: False for a in self.agents}
         self.infos = {a: {} for a in self.agents}
 
     def seed(self, seed=None):
         self.randomizer, seed = seeding.np_random(seed)
 
-    def reset(self, seed: Optional[int] = None, return_info: bool = False, options: Optional[dict] = None):
-        self.env.reset(seed=seed, return_info=return_info, options=options)
+    def reset(self, seed: Optional[int] = None, options: Optional[dict] = None):
+        self.env.reset(seed=seed, options=options)
         self.agents = self.possible_agents[:]
         self.agent_selection = self._agent_selector.reset()
         self.rewards = {agent: 0 for agent in self.agents}
         self._cumulative_rewards = {agent: 0 for agent in self.agents}
-        self.dones = {agent: False for agent in self.agents}
+        self.terminations = {a: False for a in self.agents}
+        self.truncations = {a: False for a in self.agents}
         self.infos = {agent: {} for agent in self.agents}
     
     def observation_space(self, agent):
@@ -431,8 +466,11 @@ class SumoEnvironmentPZ(AECEnv, EzPickle):
         self.env.save_csv(out_csv_name, run)
 
     def step(self, action):
-        if self.dones[self.agent_selection]:
-            return self._was_done_step(action)
+        if (
+            self.truncations[self.agent_selection]
+            or self.terminations[self.agent_selection]
+        ):
+            return self._was_dead_step(action)
         agent = self.agent_selection
         if not self.action_spaces[agent].contains(action):
             raise Exception('Action for agent {} must be in Discrete({}).'
@@ -449,7 +487,7 @@ class SumoEnvironmentPZ(AECEnv, EzPickle):
             self._clear_rewards()
         
         done = self.env._compute_dones()['__all__']
-        self.dones = {a : done for a in self.agents}
+        self.truncations = {a : done for a in self.agents}
 
         self.agent_selection = self._agent_selector.next()
         self._cumulative_rewards[agent] = 0
